/* 
   ORSA - Orbit Reconstruction, Simulation and Analysis
   Copyright (C) 2002-2003 Pasquale Tricarico
   
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifndef _ORSA_BODY_H_
#define _ORSA_BODY_H_

#include <string>
#include <list>
#include <map>
#include <iostream>

#include "orsa_coord.h"
#include "orsa_common.h"
#include "orsa_file_jpl.h"

using namespace std;

namespace orsa {
  
  class Life {
    
  public:
    inline Life() {
      born_set = dead_set = false;
    }
    
  public:
    inline void Born(UniverseTypeAwareTime &t) {
      if (dead_set) {
	cerr << "ERROR: Born called for an already dead object..." << endl;
      }
      born_set  = true;
      born_time = t;
    }
    
    inline void Dead(UniverseTypeAwareTime &t) {
      if (!born_set) {
	cerr << "ERROR: Dead called for a not yet born object..." << endl;
      }
      dead_set  = true;
      dead_time = t;
    }
    
    inline bool Born() const {
      return (born_set);
    }
    
    inline bool Dead() const {
      return (dead_set);
    }
    
    inline bool StillAlive() const {
      return (born_set && (!dead_set));
    }
    
    inline UniverseTypeAwareTime BornTime() const {
      return (born_time);
    }
    
    inline UniverseTypeAwareTime DeadTime() const {
      return (dead_time);
    }
    
    // better implementation
    // TODO: implement everything keeping into account integrations with 'dt<0'!!!
    inline bool Alive(UniverseTypeAwareTime &t) const {
      if (!born_set) return false;
      if (born_set) if (t.Time() < born_time.Time()) return false;
      if (dead_set) if (t.Time() > dead_time.Time()) return false;
      return true;
    }
    
  private:
    bool born_set, dead_set;
    UniverseTypeAwareTime born_time, dead_time;
  };
  
  class Body;
  class BodyWithEpoch;
  
  class BodyConstants {
  public:
    BodyConstants();
    BodyConstants(const string &name, const double mass);
    BodyConstants(const string &name, const double mass, const double radius);
    
    ~BodyConstants();
    
  public:
    // void InsertCopy(Body*, Body*);
    // void UpdateCopy(Body*, Body*);
    // void RemoveCopy(Body*);
    
  public:
    // void PreFree(Body*);
    // void PostAllocation(Body*);
    
  public:
    // bool IsAlive(Body*) const;
    
    //! All the bodies connected to the base body
    //  via the map_copy list.
    // void Candidates(list<Body*>&, Body* const base);
    
    inline const string name() const { return name_; } 
    inline const double mass() const { return mass_; } 
    inline const double radius() const { return radius_; }
    
  private:
    static unsigned int used_body_id;
    
  public:
    unsigned int BodyId() const { return id; }
    
  private:
    unsigned int id;
    
  protected:
    unsigned int users;
  
  public:
    void AddUser()    { ++users; }
    void RemoveUser() { --users; }
    unsigned int Users() const { return users; }	
    
  private:
    const string name_;
    const double mass_;
    const double radius_;
    // double  J2,J4; // not yet implemented
    
  protected:
    // map<Body*,Body*> map_copy;
    // map<Body*,bool>  map_alive;
    
  protected:
    static list<BodyConstants*> list_bc;
    
  protected:
    // list<Body*> list_copy;
    
    /*
      public:
      Life life;
    */
  };
  
  class Body {  
    
  public:
    // brand new bodies
    Body(); 
    Body(const double  mass);
    Body(const string &name);
    Body(const string &name, const double mass);
    Body(const string &name, const double mass, const double radius);
    Body(const string &name, const double mass, const Vector &position, const Vector &velocity);
    Body(const string &name, const double mass, const double radius, const Vector &position, const Vector &velocity);
    // copy constructor
    Body(const Body&);
    Body(const BodyWithEpoch&);
    // Body(const JPLBody&);
    // child bodies
    // Body(Body *parent);
    // Body(Body *parent, const double  mass);
    // Body(Body *parent, const string &name);
    // Body(Body *parent, const string &name, const double mass);
    // Body(Body *parent, const string &name, const double mass, const Vector &position, const Vector &velocity, bool vectors_relative_to_parent=false);
    //
    virtual ~Body();
    
    Body& operator= (const Body&);
    
  public:
    virtual string name()   const { return bc->name(); };
    virtual double mass()   const { return bc->mass(); };
    virtual double radius() const { return bc->radius(); };
    
    // inline void Born(UniverseTypeAwareTime &t) const { bc->life.Born(t); }
    // inline void Dead(UniverseTypeAwareTime &t) const { bc->life.Dead(t); }
    
    // inline bool Born() const { return (bc->life.Born()); }
    // inline bool Dead() const { return (bc->life.Dead()); }
    
    // inline bool StillAlive()                    const { return (bc->life.StillAlive());  }
    // inline bool Alive(UniverseTypeAwareTime &t) const { return (bc->life.Alive(t)); }
    
    // inline UniverseTypeAwareTime BornTime() const { return (bc->life.BornTime()); }
    // inline UniverseTypeAwareTime DeadTime() const { return (bc->life.DeadTime()); }
    
    ////
    
    inline virtual Vector position() const { 
      /* if (parent) {
	 if (bc->IsAlive(parent)) {
	 return (parent->position()+relative_position);
	 } else {
	 // WARNINGS!!!
	 }
	 } else {
	 return (relative_position);
	 }
      */
      return (_position);
    }
    
    inline virtual Vector velocity() const { 
      /* if (parent) {
	 if (bc->IsAlive(parent)) {
	 return (parent->velocity()+relative_velocity);
	 } else {
	 // WARNINGS!!!
	 }
	 } else {
         return (relative_velocity);
	 }
      */
      return (_velocity);
    }
    
    // aliases
    inline Vector Position() const { return position(); }
    inline Vector R()        const { return position(); }
    inline Vector Velocity() const { return velocity(); }
    inline Vector V()        const { return velocity(); }
    
    /*
     * NOTE: if (sync_child)
     *                       CHILD ABSOLUTE POSITION = UNCHANGED
     *                       CHILD RELATIVE POSITION = UPDATED
     *       else
     *                       CHILD ABSOLUTE POSITION = UPDATED
     *                       CHILD RELATIVE POSITION = UNCHANGED
     */                       
    
    /* 
       inline void AddToPosition(const Vector &v, bool sync_child=true) {
       relative_position += v;
       if (sync_child) {
       if (child.size()) {
       list<Body*>::iterator child_it = child.begin();
       while (child_it != child.end()) {
       (*child_it)->SyncPosition(v);
       ++child_it;
       }
       }
       }
       }
    */
    
    /* 
       inline void AddToVelocity(const Vector &v, bool sync_child=true) {
       relative_velocity += v;
       if (sync_child) {
       if (child.size()) {
       list<Body*>::iterator child_it = child.begin();
       while (child_it != child.end()) {
       (*child_it)->SyncVelocity(v);
       ++child_it;
       }
       }
       }
       }
    */
    
    inline void AddToPosition(const Vector &v) {
      _position += v;
    }
    
    inline void AddToVelocity(const Vector &v) {
      _velocity += v;
    }
    
    /* 
       inline void SetPosition(const Vector &v, bool sync_child=true, Body *ref_body=0) {
       if (ref_body) {
       
       } else {
       AddToPosition(v-position(),sync_child);
       }
       }
    */
    
    inline void SetPosition(const Vector &v) {
      _position = v;
    }
    
    /* 
       inline void SetPosition(const double x, const double y, const double z, bool sync_child=true, Body *ref_body=0) {
       const Vector v(x,y,z);
       SetPosition(v,sync_child,ref_body);
       }
    */
    
    inline void SetPosition(const double x, const double y, const double z) {
      const Vector v(x,y,z);
      SetPosition(v);
    }
    
    /* 
       inline void SetVelocity(const Vector &v, bool sync_child=true, Body *ref_body=0) {
       if (ref_body) {
       
       } else {
       AddToVelocity(v-velocity(),sync_child);
       }
       }
    */
    
    inline void SetVelocity(const double x, const double y, const double z) {
      const Vector v(x,y,z);
      SetVelocity(v);
    }
    
    /* 
       inline void SetVelocity(const Vector &v, bool sync_child=true, Body *ref_body=0) {
       if (ref_body) {
       
       } else {
       AddToVelocity(v-velocity(),sync_child);
       }
       }
    */
    
    inline void SetVelocity(const Vector &v) {
      _velocity = v;
    }
    
    // b position - this position
    Vector distanceVector(const Body &b) const;
    
    inline double distance(const Body &b) const {
      return Length(distanceVector(b));
    }
    
    // alias
    inline Vector DistanceVector(const Body &b) const { return distanceVector(b); }
    inline double Distance(const Body &b)       const { return distance(b); }
    
  public:
    unsigned int       BodyId() const { return bc->BodyId(); }
    // unsigned int ParentBodyId() const { if (parent) return parent->bc->BodyId(); }
    
    /* 
       bool HaveParent() const {
       if (parent) {
       return true;
       } else {
       return false;
       }
       }
    */
    
  public:
    // used to sort bodies by decreasing mass 
    inline bool operator < (const Body &b) const {
      return (b.mass() < mass());
    }
    
  public:
    // void SetParent(Body*, const bool soft=false, const bool recursive=false);
    
  public:
    // void RecursiveSoftSetChild(Body *old_child, Body *new_child);
    
  protected:
    //! used when the parent is changed, to update the index of the family tree 
    // void UpdateFamilyTree();
    
    /* 
       protected:
       inline void SyncPosition(const Vector &v) {
       relative_position -= v;
       }
       
       inline void SyncVelocity(const Vector &v) {
       relative_velocity -= v;
       }
    */
    
  protected:
    BodyConstants *bc;
    
  protected:
    // Vector       relative_position, relative_velocity;
    Vector _position, _velocity;
    // Body        *parent;
    // list<Body*>  child;
    
  protected:
    // static, class level family id variables
    /* 
       static unsigned int used_family_id;
       inline static unsigned int generate_family_id() { 
       return (used_family_id++);
       }
    */
    
  public:
    // inline unsigned int GetFamilyId()        const { return (family_id); };
    // inline unsigned int GetFamilyTreeLevel() const { return (family_tree_level); };
    
    friend void print(const Body&);
    
  private:
    // void local_free(const bool equal_operator=false);
    // void local_create(const Body &b, const bool equal_operator=false);
    
  protected:
    //! same id == same family tree
    // unsigned int family_id;
    
    //! 0 = base, 1 = first child....
    // unsigned int family_tree_level;
  };
  
  // better (if works....)
  /* inline bool operator== (const Body &b1, const Body &b2) {
     return (b1.BodyId() == b2.BodyId());
     } 
  */
  
  inline bool operator== (const Body &b1, const Body &b2) {
    if (b1.BodyId()             != b2.BodyId())              return false;
    // if (b1.GetFamilyId()        != b2.GetFamilyId())         return false;
    // if (b1.GetFamilyTreeLevel() != b2.GetFamilyTreeLevel())  return false;
    if (b1.name()               != b2.name())                return false;
    if (b1.mass()               != b2.mass())                return false;
    if (b1.position()           != b2.position())            return false;
    if (b1.velocity()           != b2.velocity())            return false;
    return true;
  } 
  
  /* 
     inline bool operator== (const Body &b1, const Body &b2) {
     if (b1.GetFamilyId()        != b2.GetFamilyId())         return false;
     if (b1.GetFamilyTreeLevel() != b2.GetFamilyTreeLevel())  return false;
     if (b1.name()               != b2.name())                return false;
     if (b1.mass()               != b2.mass())                return false;
     if (b1.position()           != b2.position())            return false;
     if (b1.velocity()           != b2.velocity())            return false;
     return true;
     }
  */
  
  inline bool operator!= (const Body &b1, const Body &b2) {
    return !(b1 == b2);
  }
  
  //! base element for intepolation
  class BodyWithParameter : public Body {
  public:
    double par;
  };
  
  class BodyWithEpoch : public Body {
  public:
    inline BodyWithEpoch() : Body() { }
    
  public:
    inline BodyWithEpoch(const Body &b) : Body(b) { }
    inline BodyWithEpoch(Body *b) : Body(b) { }
			  
    inline BodyWithEpoch(const double mass) : Body(mass) { }
			  
    inline BodyWithEpoch(const string &name, const double mass) : Body(name,mass) { }
			  
    inline BodyWithEpoch(const string &name, const double mass, const Vector &r, const Vector &v) : Body(name,mass,r,v) { }
     
    inline BodyWithEpoch(const string &name, const double mass, const Vector &r, const Vector &v, const Date &d) : Body(name,mass,r,v) {
      epoch.SetDate(d);
    }
    
  public:
    UniverseTypeAwareTime epoch;
  };
  
  
  // JPL planets
  class JPLBody : public BodyWithEpoch {
  public:
    JPLBody(JPL_planets p, const Date &epoch);
    JPLBody(const JPLBody&);
    ~JPLBody();
    
  public:
    void SetEpoch(const Date &epoch);
    
  public:
    string name()   const;
    double mass()   const;
    double radius() const;
    
  public:
    Vector position() const;
    Vector velocity() const;
    
  public:
    JPL_planets JPLPlanet() const { return planet; }
    Date        Epoch() const { return epoch; }
    
  private:
    JPLFile     *jf;
    JPL_planets  planet;
    Date         epoch;
    Vector       r; 
    Vector       v;
  };
  
  void Interpolate(const vector<BodyWithParameter> &b_in, const double x, Body &b_out, Body &err_b_out);
  
  void print(const Body&);
  
  double KineticEnergy(const Body&);
  
  // template <class T> void AutoHierarchy(vector<T>&); // T is usually a Body or a BodyWithEpoch
  
  // template <class T> void PrintHierarchy(const vector<T>&); // T is usually a Body or a BodyWithEpoch
  
} // namespace orsa

#endif // _ORSA_BODY_H_
