/* 
   ORSA - Orbit Reconstruction, Simulation and Analysis
   Copyright (C) 2002-2003 Pasquale Tricarico
   
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifndef _ORSA_COORD_H_
#define _ORSA_COORD_H_

#include <vector>

namespace orsa {
  
  using std::vector;
  
  class Vector {
    
  public:
    
    // constructors
    inline Vector() { x = y = z = 0; }
   
    inline Vector(const Vector& v) {
      x = v.x;
      y = v.y;
      z = v.z;
    }
    
    inline Vector(double _x, double _y, double _z) {
      x = _x;
      y = _y;
      z = _z;
    }
    
    // utils
    inline double GetX() const { return x; }
    inline double GetY() const { return y; }
    inline double GetZ() const { return z; }
    
    // operators
    inline Vector& operator+= (const Vector& v) {
      x += v.x;
      y += v.y;
      z += v.z;
      return *this;
    }
    
    inline Vector& operator-= (const Vector& v) {
      x -= v.x;
      y -= v.y;
      z -= v.z;
      return *this;
    }
    
    inline Vector& operator*= (const double f) {
      x *= f;
      y *= f;
      z *= f;
      return *this;
    }
    
    inline Vector& operator/= (const double f) {
      x /= f;
      y /= f;
      z /= f;
      return *this;
    }
    
    // sign
    inline Vector operator+ () {
      return Vector(x,y,z);
    }
    
    inline Vector operator- () {
      return Vector(-x,-y,-z);
    }
    
    inline Vector operator* (const double f) {
      return Vector(x*f, y*f, z*f);
    }
    
    inline Vector operator/ (const double f) {
      return Vector(x/f, y/f, z/f);
    }
    
    // rotation
    Vector& rotate (const double, const double, const double);
    
    // set
    inline void Set(Vector v) {
      x = v.x;
      y = v.y;
      z = v.z;
    }
    
    inline void Set(double _x, double _y, double _z) {
      x = _x;
      y = _y;
      z = _z;
    }
    
  public:
    double x, y, z;
  };
  
  inline Vector operator* (const double f, const Vector& v) {
    return Vector(v.x*f, v.y*f, v.z*f);
  }
  
  inline Vector operator* (const Vector& v, const double f) {
    return Vector(v.x*f, v.y*f, v.z*f);
  }
  
  inline Vector operator/ (const Vector& v, const double f) {
    return Vector(v.x/f, v.y/f, v.z/f);
  }
  
  inline Vector operator+ (const Vector& u, const Vector& v) {
    return Vector(u.x+v.x,
                  u.y+v.y,
                  u.z+v.z);
  }
  
  inline Vector operator- (const Vector& u, const Vector& v) {
    return Vector(u.x-v.x,
                  u.y-v.y,
                  u.z-v.z);
  }
  
  inline Vector ExternalProduct (const Vector& u, const Vector& v) {
    return Vector (u.y*v.z-u.z*v.y,
		   u.z*v.x-u.x*v.z,
		   u.x*v.y-u.y*v.x); 
  }
  
  inline Vector Cross (const Vector& u, const Vector& v) {
    return Vector (u.y*v.z-u.z*v.y,
		   u.z*v.x-u.x*v.z,
		   u.x*v.y-u.y*v.x); 
  }
  
  // scalar product
  inline double operator* (const Vector& u, const Vector& v) {
    return (u.x*v.x+
	    u.y*v.y+
	    u.z*v.z);
  }
  
  double Length(const Vector&);
  
  inline double LengthSquared(const Vector &v) {
    return ( (v.x*v.x) +
             (v.y*v.y) +
             (v.z*v.z) );
  }
  
  inline bool operator== (const Vector &v1, const Vector &v2) {
    if (v1.x != v2.x) return false;
    if (v1.y != v2.y) return false;
    if (v1.z != v2.z) return false;
    return true;
  }
  
  inline bool operator!= (const Vector &v1, const Vector &v2) {
    return !(v1 == v2);
  }
  
  // container for intepolation
  class VectorWithParameter : public Vector {
  public:
    double par;
  };
  
  void Interpolate(const vector<VectorWithParameter> v_in, const double x, Vector &v_out, Vector &err_v_out);
  
} // namespace orsa

#endif // _ORSA_COORD_H_
