/* 
   ORSA - Orbit Reconstruction, Simulation and Analysis
   Copyright (C) 2002-2003 Pasquale Tricarico
   
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifndef _ORSA_INTERACTION_H_
#define _ORSA_INTERACTION_H_

#include <vector>

#include "orsa_frame.h"
#include "orsa_coord.h"
#include "orsa_common.h"

namespace orsa {
  
  // == Unique id == if you need to modify this list, don't cnahge the numbers, but add new numbers
  enum InteractionType {
    NEWTON=9001,
    ARMONICOSCILLATOR=9002,
    GALACTIC_POTENTIAL_ALLEN=9003,
    GALACTIC_POTENTIAL_ALLEN_PLUS_NEWTON=9004,
    JPL_PLANETS_NEWTON=9005
  };
   
  inline string label(const InteractionType it) {
    
    string s;
    
    switch (it) {
    case NEWTON:                                s = "Newton";                               break;
    case ARMONICOSCILLATOR:                     s = "Armonic Oscillator";                   break;
    case GALACTIC_POTENTIAL_ALLEN:              s = "Galactic Potential (Allen)";           break; 
    case GALACTIC_POTENTIAL_ALLEN_PLUS_NEWTON:  s = "Galactic Potential (Allen) + Newton";  break;
    case JPL_PLANETS_NEWTON:                    s = "JPL planets + Newton";                 break; 
    }	
    
    return s;
  }
  
  // TODO: distinguish between conservative and non-conservative interactions
  
  // base class Interaction
  
  class Interaction {
  public:
    virtual ~Interaction() {};
  public:
    virtual void Acceleration(const Frame&, vector<Vector>&) = 0;  
    virtual double PotentialEnergy(const Frame&) = 0;
    
  public:
    InteractionType type;
  };
  
  // derived classes
  
  class Newton : public Interaction {
  public:
    Newton();
    
    void Acceleration(const Frame&, vector<Vector>&);
    double PotentialEnergy(const Frame&);
    
  private:
    double g;
  };
  
  class ArmonicOscillator : public Interaction {
  public:
    ArmonicOscillator(const double free_length_in, const double k_in);
    
    void Acceleration(const Frame&, vector<Vector>&);
    double PotentialEnergy(const Frame&);
    
  private:
    double free_length;
    double k; // 'spring' constant
  };
  
  class GalacticPotentialAllen : public Interaction {
  public:
    GalacticPotentialAllen();
    
    void Acceleration(const Frame&, vector<Vector>&);
    double PotentialEnergy(const Frame&);
    
  private:
    double g;
    double mb; // bulge mass
    double bb; // bulge length scale
    double md; // disk mass
    double ad; // disk length scale
    double bd; // disk height scale
    double mh; // halo mass (different from the reference article, rescaled to obtain a circular velocity at the sun radius of about 220 km/s)
    double ah; // halo length scale
  };
  
  class GalacticPotentialAllenPlusNewton : public Interaction {
  public:
    GalacticPotentialAllenPlusNewton() {
      type = GALACTIC_POTENTIAL_ALLEN_PLUS_NEWTON; 
    }
    
    inline void Acceleration(const Frame &f, vector<Vector> &a) {
      
      tmp_a.resize(a.size());
      
      unsigned int i;
      
      for (i=0;i<a.size();++i)
	a[i].Set(0,0,0);
      
      gpa_itg.Acceleration(f,tmp_a);
      for (i=0;i<a.size();++i)
	a[i] += tmp_a[i];
      
      newton_itg.Acceleration(f,tmp_a);
      for (i=0;i<a.size();++i)
	a[i] += tmp_a[i];
      
    }	
    
    inline double PotentialEnergy(const Frame &f) {
      return (gpa_itg.PotentialEnergy(f)+newton_itg.PotentialEnergy(f));
    }
    
  private:
    GalacticPotentialAllen    gpa_itg;
    Newton                 newton_itg;
    //
    vector<Vector> tmp_a;
  };
   
  class JPLPlanetsNewton : public Interaction {
  public:
    JPLPlanetsNewton(list<JPL_planets>&);
    
    void Acceleration(const Frame&, vector<Vector>&);
    double PotentialEnergy(const Frame&);
    
  private:
    Newton newton_itg;
    list<JPL_planets> l;
    Frame planets_frame;
    double g;
  };
  
} // namespace orsa

#endif // _ORSA_INTERACTION_H_
