/* 
   ORSA - Orbit Reconstruction, Simulation and Analysis
   Copyright (C) 2002-2003 Pasquale Tricarico
   
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifndef _ORSA_ORBIT_H_
#define _ORSA_ORBIT_H_

#include "orsa_coord.h"
#include "orsa_universe.h"
#include "orsa_frame.h"
#include "orsa_analysis.h"

#include <string>
#include <vector>

using namespace std;

namespace orsa {
  
  class Observation {
  public:
    // used to sort observation 
    inline bool operator < (const Observation &obs) const {
      return (date.GetJulian() < obs.date.GetJulian());
    }
    
  public:
    string designation;
    string discovery_asterisk;
    Date   date;
    Angle  ra, dec;         
    double mag; // magnitude
    // unsigned int obscode;
    string obscode;
  };
  
  class Sky {
  public:
    void Set(Angle &ra, Angle &dec, UniverseTypeAwareTime &epoch) {
      _ra    = ra;
      _dec   = dec;
      _epoch = epoch;
    }
    
  public:
    inline Angle ra()  const { return _ra; }
    inline Angle dec() const { return _dec; }
    inline UniverseTypeAwareTime epoch() const { return _epoch; }
    
  public:
    void Compute(const Vector &relative_position, const UniverseTypeAwareTime& epoch);
    
  public:
    double delta_arcsec(const Observation &obs) const;
    
  private:
    Angle _ra, _dec;
    UniverseTypeAwareTime _epoch; // the epoch is used only to obtain the obleq angle
  };
  
  
  class Orbit {
  public:
    Orbit() { };
    
  public:
    inline double Period() const { return (sqrt(4*pisq/mu*pow(a,3))); }
    
  public:
    void RelativePosVel(Vector &relative_position, Vector &relative_velocity) const;
    
  public:
    void Compute(const Vector &dr, const Vector &dv, const double mu);
    void Compute(const Body &b, const Body &ref_b);
    
  public:
    double a,e,i,omega_node,omega_pericenter,M;
    double mu; // G*(m+M)
  };
  
  class OrbitWithEpoch : public Orbit {
  public:
    inline OrbitWithEpoch() : Orbit() { }
    inline OrbitWithEpoch(const Orbit &orbit) : Orbit(orbit) { }
    inline OrbitWithEpoch(const OrbitWithEpoch &orbit) : Orbit(orbit), epoch(orbit.epoch) { }
    		   
  public:
    inline void Compute(const Vector &dr, const Vector &dv, const double mu) {
      Orbit::Compute(dr,dv,mu);
    }
    
    inline void Compute(const Vector &dr, const Vector &dv, const double mu_in, const UniverseTypeAwareTime &epoch_in) {
      epoch = epoch_in;
      Orbit::Compute(dr,dv,mu);
    }
    
    inline void Compute(const Body &b, const Body &ref_b) {
      Orbit::Compute(b, ref_b);
    }
    
    inline void Compute(const Body &b, const Body &ref_b, const UniverseTypeAwareTime &epoch_in) {
      epoch = epoch_in;
      Orbit::Compute(b, ref_b);
    }
    
  public:    
    inline void RelativePosVel(Vector &relative_position, Vector &relative_velocity) const {
      Orbit::RelativePosVel(relative_position,relative_velocity);
    }
    
  public:
    inline void RelativePosVel(Vector &relative_position, Vector &relative_velocity, const UniverseTypeAwareTime &epoch_in) const {
      OrbitWithEpoch o(*this);
      o.M += twopi*(epoch_in.Time() - epoch.Time())/Period();
      o.M  = fmod(10*twopi+fmod(o.M,twopi),twopi);
      o.RelativePosVel(relative_position,relative_velocity);
    }
    
  public:
    UniverseTypeAwareTime epoch;
    double libration_angle; // to be removed soon!!
  };
  
  //! The RMS of the residuals in arcsec units
  double RMS_residuals(const vector<Observation>&, const OrbitWithEpoch&);
  
  enum CovarianceMatrixElements { Osculating, // a,e,i,node,peri,M
				  Equinoctal  // a,k,h,q,p,lambda
  };
  
  class OrbitWithCovarianceMatrix : public OrbitWithEpoch {
  public:
    OrbitWithCovarianceMatrix(Orbit&,double**,CovarianceMatrixElements);
    
  private:
    double covm[6][6];
    CovarianceMatrixElements cov_base;
  public:
    void generate(vector<Orbit>&,const int);
  };
  
  class OrbitStream : public vector<OrbitWithEpoch> {
  public:    
    // double window_amplitude,window_start,window_step;
    OrbitStream(): label("") {}
    
  public:
    int               asteroid_number;  
    double            timestep;
    WindowParameters  wp;    
    string            label; // used in output file names
  };
  
  Vector Position(const Orbit&);
  Vector Position(const Orbit&, const double);
  
  Vector Velocity(const Orbit&);
  Vector Velocity(const Orbit&, const double);
  
  void RelativePosVel(const Orbit&, Vector&, Vector&);
  void RelativePosVel(const Orbit&, Vector&, Vector&, const double);
  void RelativePosVel(const Orbit&, Vector&, Vector&, const Date);
  
  // sky position of an object
  Sky PropagatedPosition(const OrbitWithEpoch&, const UniverseTypeAwareTime&, const string &obscode, bool integrate=true);
  
  // an optimized class for frequent position computations of an object with a given orbit
  class OptimizedOrbitPositions {
  public:
    OptimizedOrbitPositions(const OrbitWithEpoch &orbit);
    
  public:
    Sky PropagatedPosition(const UniverseTypeAwareTime&, const string &obscode, bool integrate=true);
    OrbitWithEpoch PropagatedOrbit(const UniverseTypeAwareTime&, bool integrate=true);
    
  private:
    const OrbitWithEpoch _orbit;
    // Config conf;
    // LocationFile lf;
    list<JPL_planets> l;
    vector<Frame> frames;
  };
  
  class Asteroid {
  public:
    int n;    
    OrbitWithEpoch orb;
    string name;
    double mag;
    double ceu;
  };
  
  class AsteroidDatabase : public vector<Asteroid> {
    
  };
  
  class ObservationCandidate { 
  public:
    Asteroid        a;
    Observation     o;
    Angle           delta;
  };
  
  class Location {
  public:
    double lon,pxy,pz;
    string name;
    bool   present;  
  };
  
} // namespace orsa

#endif // _ORSA_ORBIT_H_
