/* 
   ORSA - Orbit Reconstruction, Simulation and Analysis
   Copyright (C) 2002-2003 Pasquale Tricarico
   
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifndef _ORSA_UNITS_H
#define _ORSA_UNITS_H

#include <cmath>
#include <string>

// #include "orsa_common.h" // Don't include "orsa_common.h" to avoid a loop in inclusions...

using namespace std;

namespace orsa {
  
  using namespace std;
  
  enum time_unit   {YEAR=0,DAY,HOUR,MINUTE,SECOND};
  
  enum length_unit {MPARSEC=98,KPARSEC=99,PARSEC=100,LY,AU,EARTHMOON,REARTH,RMOON,KM,M,CM};
  
  enum mass_unit   {MSUN=200,MJUPITER,MEARTH,MMOON,KG,GRAM};
  
  // const double G_MKS = 6.67259e-11;
  extern const double G_MKS;
  const double     MSUN_MKS = 1.9889e30;
  const double MJUPITER_MKS = 1.8989e27;
  const double   MEARTH_MKS = 5.9742e24;
  const double    MMOON_MKS = 7.3483e22;
  const double       AU_MKS = 1.49597870660e11;
  const double        c_MKS = 299792458.0;
  const double  R_EARTH_MKS = 6378140.0;
  const double   R_MOON_MKS = 1737400.0;
  
  template <class UNIT> class UnitBaseScale {
  public:
    UnitBaseScale() {}
    UnitBaseScale(UNIT unit) { base_unit = unit; }
  public:
    inline void   Set(UNIT unit) { base_unit = unit; }
    inline UNIT   GetBaseUnit() const { return base_unit; }
    
  private:
    UNIT base_unit;
  };
  
  class Units {
    
  private:
    UnitBaseScale<time_unit>   Time;
    UnitBaseScale<length_unit> Length;
    UnitBaseScale<mass_unit>   Mass;
    
  public:
    Units();
    Units(time_unit, length_unit, mass_unit);
    
  private:
    void init_base();
    
  private:
    void TryToSetUnitsFromJPLFile();
    
  public:
    void SetSystem(time_unit tu, length_unit lu, mass_unit mu);
    
    // constants
    inline double GetG()    const { return G; };
    inline double GetMSun() const { return MSun; };
    
  public:
    string label(time_unit)   const;
    string label(length_unit) const;
    string label(mass_unit)   const;
    
  public:
    inline string TimeLabel()   const { return label(  Time.GetBaseUnit()); };
    inline string LengthLabel() const { return label(Length.GetBaseUnit()); };
    inline string MassLabel()   const { return label(  Mass.GetBaseUnit()); };
    
    // conversions
    // new version -- put a world-visible version outside this class but inside the orsa namespace!
  public:
    inline double FromUnits(double x, time_unit   t_in, int power=1) const { return (x*pow(GetTimeScale(t_in)/GetTimeScale(),power));     }
    inline double FromUnits(double x, length_unit l_in, int power=1) const { return (x*pow(GetLengthScale(l_in)/GetLengthScale(),power)); }
    inline double FromUnits(double x, mass_unit   m_in, int power=1) const { return (x*pow(GetMassScale(m_in)/GetMassScale(),power));     }
    
  public:
    inline time_unit   GetTimeBaseUnit()   const { return Time.GetBaseUnit(); };
    inline length_unit GetLengthBaseUnit() const { return Length.GetBaseUnit(); };
    inline mass_unit   GetMassBaseUnit()   const { return Mass.GetBaseUnit(); };
    
  protected:
    double GetTimeScale(  const   time_unit tu) const; 
    double GetLengthScale(const length_unit lu) const; 
    double GetMassScale(  const   mass_unit mu) const; 
    
    double GetTimeScale()   const;
    double GetLengthScale() const; 
    double GetMassScale()   const; 
    
    void Recompute();
    
  private:
    double G,G_base;
    double MSun,MSun_base;
    double MJupiter_base, MEarth_base, MMoon_base;
    double AU_base;
    double c_base;
    double r_earth_base;
    double r_moon_base;
    double parsec_base;
  };
  
  // defined in orsa_universe.cc
  double GetG();
  double GetMSun();
  
  /* double FromUnits(double, time_unit);
     double FromUnits(double, length_unit);
     double FromUnits(double, mass_unit); 
  */
  
  // everything in orsa_universe.cc
  double FromUnits(double, time_unit,   int = 1);
  double FromUnits(double, length_unit, int = 1);
  double FromUnits(double, mass_unit,   int = 1); 
  //
  string TimeLabel();
  string LengthLabel();
  string MassLabel();
  
  //! TimeScale enum, useful only when using a Real Universe. More informations can be obtained here: http://www.hartrao.ac.za/nccsdoc/slalib/sun67.htx/node217.html
  enum TimeScale {UTC,UT,UT1=UT,TAI,TDT,ET=TDT,TT=TDT,GPS};
  
  typedef struct TAI_minus_UTC_element {
    int day,month,year;
    int TAI_minus_UTC;
  };
  
  inline bool operator== (const TAI_minus_UTC_element &x, const TAI_minus_UTC_element &y) {
    if (x.day   != y.day)   return false;
    if (x.month != y.month) return false;
    if (x.year  != y.year)  return false;
    if (x.TAI_minus_UTC != y.TAI_minus_UTC) return false;
    return true;
  }
  
  inline bool operator!= (const TAI_minus_UTC_element &x, const TAI_minus_UTC_element &y) {
    return !(x==y);
  }
  
  /*
   *  The TAI_minus_UTC_table should be updated periodically, 
   * for instance by consulting:
   * 
   * http://hpiers.obspm.fr/webiers/general/communic/publi/PUBLI.html
   * 
   * The present version is updated through Bulletin C17 (Jan 28, 1999)
   * Day, month, year, TAI-UTC (s)
   */
  
  const TAI_minus_UTC_element TAI_minus_UTC_table_final_element = {0,0,0,0};
  const TAI_minus_UTC_element TAI_minus_UTC_table[] = {
    {1,1,1972,10},
    {1,7,1972,11},
    {1,1,1973,12},
    {1,1,1974,13},
    {1,1,1975,14},
    {1,1,1976,15},
    {1,1,1977,16},
    {1,1,1978,17},
    {1,1,1979,18},
    {1,1,1980,19},
    {1,7,1981,20},
    {1,7,1982,21},
    {1,7,1983,22},
    {1,7,1985,23},
    {1,1,1988,24},
    {1,1,1990,25},
    {1,1,1991,26},
    {1,7,1992,27},
    {1,7,1993,28},
    {1,7,1994,29},
    {1,1,1996,30},
    {1,7,1997,31},
    {1,1,1999,32}, // latest leap second, updated Jan 2003
    TAI_minus_UTC_table_final_element
  };
  
  typedef struct ET_minus_UT_element {
    int day,month,year;
    double ET_minus_UT;
  };
  
  inline bool operator== (const ET_minus_UT_element &x, const ET_minus_UT_element &y) {
    if (x.day   != y.day)   return false;
    if (x.month != y.month) return false;
    if (x.year  != y.year)  return false;
    if (x.ET_minus_UT != y.ET_minus_UT) return false;
    return true;
  }
  
  inline bool operator!= (const ET_minus_UT_element &x, const ET_minus_UT_element &y) {
    return !(x==y);
  }
  
  /* 
   * Values ET - UT at 0h UT of the date (specified as day, month, year)
   * from The Astronomical Almanac 1999, pages K8-9
   */
  
  const ET_minus_UT_element ET_minus_UT_table_final_element = {0,0,0,0};
  const ET_minus_UT_element ET_minus_UT_table[] = {
    {1,1,1800,13.7},
    {1,1,1801,13.4},
    {1,1,1802,13.1},
    {1,1,1803,12.9},
    {1,1,1804,12.7},
    {1,1,1805,12.6},
    {1,1,1806,12.5},
    {1,1,1816,12.5},
    {1,1,1817,12.4},
    {1,1,1818,12.3},
    {1,1,1819,12.2},
    {1,1,1820,12.0},
    {1,1,1821,11.7},
    {1,1,1822,11.4},
    {1,1,1823,11.1},
    {1,1,1824,10.6},
    {1,1,1825,10.2},
    {1,1,1826, 9.6},
    {1,1,1827, 9.1},
    {1,1,1828, 8.6},
    {1,1,1829, 8.0},
    {1,1,1830, 7.5},
    {1,1,1831, 7.0},
    {1,1,1832, 6.6},
    {1,1,1833, 6.3},
    {1,1,1834, 6.0},
    {1,1,1835, 5.8},
    {1,1,1836, 5.7},
    {1,1,1837, 5.6},
    {1,1,1838, 5.6},
    {1,1,1839, 5.6},
    {1,1,1840, 5.7},
    {1,1,1841, 5.8},
    {1,1,1842, 5.9},
    {1,1,1843, 6.1},
    {1,1,1844, 6.2},
    {1,1,1845, 6.3},
    {1,1,1846, 6.5},
    {1,1,1847, 6.6},
    {1,1,1848, 6.8},
    {1,1,1849, 6.9},
    {1,1,1850, 7.1},
    {1,1,1851, 7.2},
    {1,1,1852, 7.3},
    {1,1,1853, 7.4},
    {1,1,1854, 7.5},
    {1,1,1855, 7.6},
    {1,1,1856, 7.7},
    {1,1,1857, 7.7},
    {1,1,1858, 7.8},
    {1,1,1859, 7.8},
    {1,1,1860, 7.88},
    {1,1,1861, 7.82},
    {1,1,1862, 7.54},
    {1,1,1863, 6.97},
    {1,1,1864, 6.40},
    {1,1,1865, 6.02},
    {1,1,1866, 5.41},
    {1,1,1867, 4.10},
    {1,1,1868, 2.92},
    {1,1,1869, 1.82},
    {1,1,1870, 1.61},
    {1,1,1871, 0.10},
    {1,1,1872,-1.02},
    {1,1,1873,-1.28},
    {1,1,1874,-2.69},
    {1,1,1875,-3.24},
    {1,1,1876,-3.64},
    {1,1,1877,-4.54},
    {1,1,1878,-4.71},
    {1,1,1879,-5.11},
    {1,1,1880,-5.40},
    {1,1,1881,-5.42},
    {1,1,1882,-5.20},
    {1,1,1883,-5.46},
    {1,1,1884,-5.46},
    {1,1,1885,-5.79},
    {1,1,1886,-5.63},
    {1,1,1887,-5.64},
    {1,1,1888,-5.80},
    {1,1,1889,-5.66},
    {1,1,1890,-5.87},
    {1,1,1891,-6.01},
    {1,1,1892,-6.19},
    {1,1,1893,-6.64},
    {1,1,1894,-6.44},
    {1,1,1895,-6.47},
    {1,1,1896,-6.09},
    {1,1,1897,-5.76},
    {1,1,1898,-4.66},
    {1,1,1899,-3.74},
    {1,1,1900,-2.72},
    {1,1,1901,-1.54},
    {1,1,1902,-0.02},
    {1,1,1903, 1.24},
    {1,1,1904, 2.64},
    {1,1,1905, 3.86},
    {1,1,1906, 5.37},
    {1,1,1907, 6.14},
    {1,1,1908, 7.75},
    {1,1,1909, 9.13},
    {1,1,1910,10.46},
    {1,1,1911,11.53},
    {1,1,1912,13.36},
    {1,1,1913,14.65},
    {1,1,1914,16.01},
    {1,1,1915,17.20},
    {1,1,1916,18.24},
    {1,1,1917,19.06},
    {1,1,1918,20.25},
    {1,1,1919,20.95},
    {1,1,1920,21.16},
    {1,1,1921,22.25},
    {1,1,1922,22.41},
    {1,1,1923,23.03},
    {1,1,1924,23.49},
    {1,1,1925,23.62},
    {1,1,1926,23.86},
    {1,1,1927,24.49},
    {1,1,1928,24.34},
    {1,1,1929,24.08},
    {1,1,1930,24.02},
    {1,1,1931,24.00},
    {1,1,1932,23.87},
    {1,1,1933,23.95},
    {1,1,1934,23.86},
    {1,1,1935,23.93},
    {1,1,1936,23.73},
    {1,1,1937,23.92},
    {1,1,1938,23.96},
    {1,1,1939,24.02},
    {1,1,1940,24.33},
    {1,1,1941,24.83},
    {1,1,1942,25.30},
    {1,1,1943,25.70},
    {1,1,1944,26.24},
    {1,1,1945,26.77},
    {1,1,1946,27.28},
    {1,1,1947,27.78},
    {1,1,1948,28.25},
    {1,1,1949,28.71},
    {1,1,1950,29.15},
    {1,1,1951,29.57},
    {1,1,1952,29.97},
    {1,1,1953,30.36},
    {1,1,1954,30.72},
    {1,1,1955,31.07},
    {1,1,1956,31.35},
    {1,1,1957,31.68},
    {1,1,1958,32.18},
    {1,1,1959,32.68},
    {1,1,1960,33.15},
    {1,1,1961,33.59},
    {1,1,1962,34.00},
    {1,1,1963,34.47},
    {1,1,1964,35.03},
    {1,1,1965,35.73},
    {1,1,1966,36.54},
    {1,1,1967,37.43},
    {1,1,1968,38.29},
    {1,1,1969,39.20},
    {1,1,1970,40.18},
    {1,1,1971,41.17},
    {1,1,1972,42.23},
    {1,1,1973,43.37},
    {1,1,1974,44.49},
    {1,1,1975,45.48},
    {1,1,1976,46.46},
    {1,1,1977,47.52},
    {1,1,1978,48.53},
    {1,1,1979,49.59},
    {1,1,1980,50.54},
    {1,1,1981,51.38},
    {1,1,1982,52.17},
    {1,1,1983,52.96},
    {1,1,1984,53.79},
    {1,1,1985,54.34},
    {1,1,1986,54.87},
    {1,1,1987,55.32},
    {1,1,1988,55.82},
    {1,1,1989,56.30},
    {1,1,1990,56.86},
    {1,1,1991,57.57},
    {1,1,1992,58.31},
    {1,1,1993,59.12},
    {1,1,1994,59.98},
    {1,1,1995,60.78},
    {1,1,1996,61.63},
    {1,1,1997,62.29},
    {1,1,1998,62.97},
    {1,1,1999,63.46},
    {1,1,2000,63.83},
    {1,1,2001,64.09},
    {1,1,2002,64.30},
    {1,1,2003,64.47},
    ET_minus_UT_table_final_element
  };
  
  const TimeScale default_Date_timescale = ET;
  
  string TimeScaleLabel(TimeScale);
  
  // double delta_seconds(const TimeScale to, const TimeScale from); // time in 'to' = time in 'from' + delta_seconds
  
  class UniverseTypeAwareTime;
  
  class Date {
  public:
    inline Date() : scale(default_Date_timescale) {};
    inline Date(TimeScale ts) : scale(ts) {};
    Date(UniverseTypeAwareTime &t);
    
  public:
    void SetGregor(int y, int m, double d);
    void GetGregor(int &y, int &m, int &d) const;
    //
    void   SetJulian(double);
    void   GetJulian(double&) const;
    double GetJulian() const;
    // 
    double GetDayFraction() const { return frac; }
    //
    double GetTime() const;
    void SetTime(const double);
    
  public:
    void SetNow();
    
  public:
    //! changes the timescale without converting the date
    void SetTimeScale(TimeScale ts) { scale=ts; };
    TimeScale GetTimeScale() const { return scale; };
    
  public:
    void ConvertToTimeScale(const TimeScale ts=default_Date_timescale);
    
  private:
    long int  sdn;  // internal day format
    double    frac; /* fraction of day:  0h -> frac = 0.0;
		     *                  12h -> frac = 0.5; 
		     */
    
    TimeScale scale;
  };
  
  inline bool operator== (const Date &x, const Date &y) {
    if (x.GetJulian()    != y.GetJulian())    return false;
    if (x.GetTimeScale() != y.GetTimeScale()) return false;
    return true;
  }
  
  inline bool operator!= (const Date &x, const Date &y) {
    return !(x==y);
  }
  
  // UniverseTypeAwareTime
  
  //! A Date like class which represents time
  //! as Date in Real Universes and as a double
  //! in Simulated Universes  
  class UniverseTypeAwareTime {
  public:
    UniverseTypeAwareTime();
    UniverseTypeAwareTime(const double);
    UniverseTypeAwareTime(const Date&);
    UniverseTypeAwareTime(const UniverseTypeAwareTime&);
    
  public:
    double Time()    const;
    double GetTime() const;
    Date   GetDate() const;
    void   SetTime(const double);
    void   SetTime(const Date &);
    void   SetDate(const Date &);
    
  protected:
    double time;
    Date   date;
  };
  
  // Angle
  
  class Angle {
  public:
    void   SetRad(double);
    void   GetRad(double&) const;
    double GetRad() const;
    //
    void   SetDPS(double  d, double  p, double  s);
    void   GetDPS(double &d, double &p, double &s) const;
    //
    void   SetHMS(double  h, double  m, double  s);
    void   GetHMS(double &h, double &m, double &s) const;

  private:
    double radiants;
  };
  
  using std::sin;
  using std::cos; 
  using std::tan;
  
  inline double sin(const Angle &alpha) {
    return sin(alpha.GetRad());
  }
  
  inline double cos(const Angle &alpha) {
    return cos(alpha.GetRad());
  }
  
  inline double tan(const Angle &alpha) {
    return tan(alpha.GetRad());
  }
  
  // reference systems
  
  enum ReferenceSystem {EQUATORIAL=8001,ECLIPTIC=8002};
  
  Angle obleq(const Date);
  
  Angle gmst(const Date);
  
} // namespace orsa

#endif // _ORSA_UNITS_H
