/* 
   ORSA - Orbit Reconstruction, Simulation and Analysis
   Copyright (C) 2002-2003 Pasquale Tricarico
   
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifndef _ORSA_UNIVERSE_H_
#define _ORSA_UNIVERSE_H_

#include <string>
#include <vector>

#include "orsa_body.h"
#include "orsa_integrator.h"
#include "orsa_units.h"
#include "orsa_common.h"
#include "orsa_frame.h"

using namespace std;

namespace orsa {
  
  //! This class collects all the Frame's of an integration, sampled at a fixed sample_period
  class Evolution : public vector<Frame> {
  public:
    Evolution();
    virtual ~Evolution();
    
  public:
    void Integrate(const UniverseTypeAwareTime &time_stop);
    
  public:
    inline virtual void step_done(double,double,double,Frame&,bool&) { };
    inline virtual void integration_finished() { };
    
  public:
    inline void SetMaxUnsavedSubSteps(int m) {
      max_unsaved_substeps = m;
      max_unsaved_substeps_active = true;
    }
    
  public:
    vector<BodyWithEpoch> start_bodies;
    vector<JPLBody>       start_JPL_bodies; // used only in Real Universe mode
    
  public:
    Integrator    *integrator;
    Interaction   *interaction;
    double         sample_period;
    
  private:
    int  max_unsaved_substeps;
    bool max_unsaved_substeps_active;
    
  public:
    string  name;
  };
  
  //! This enum is used to classify the Universes: a Real Universe is composed exclusively by the JPL planets and massless bodies, while the Simulated Universe is composed by an arbitrary number of massive and massless objects. Another big difference is the following one: in the Real Universe each body can be defined at a different epoch, while in the Simulated all the bodies are defined at the same time.
  enum UniverseType {Real=9001,Simulated=9002};
  
  class Universe : public vector<Evolution*> { 
  public:
    
    Universe();
    Universe(length_unit,mass_unit,time_unit,ReferenceSystem=ECLIPTIC,UniverseType=Simulated);
    
    ~Universe();
    
  public:
    inline ReferenceSystem GetReferenceSystem() const { return sys; }
    inline UniverseType GetUniverseType() const { return type; }
    
  public:
    string  name;
    string  description;
    
  public:
    bool    modified;
    
  private:
    ReferenceSystem sys;
    UniverseType    type;
  };
  
  //! The active universe
  extern Universe *universe;
  
  //! A good frame to start an integration with 
  // template <class T> Frame StartFrame(const vector<T>&, Interaction*, Integrator*, const UniverseTypeAwareTime&); // T is usually a BodyWithEpoch
  Frame StartFrame(const vector<BodyWithEpoch>&, vector<JPLBody>&, Interaction*, Integrator*, const UniverseTypeAwareTime&);
  
} // namespace orsa

#endif // _ORSA_UNIVERSE_H_
